#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <math.h>
#include <limits.h>

#include "wavfuncs.h"

/**

wavfade - a tool to fade-in/out 16 bit stereo .wav files 

duncan@zog.net.au 19/7/1999,
frederik.arbeiter@gmx.net June 2000

$Id: wavfade.c,v 1.5 1999/07/28 13:18:06 duncan Exp $

**/



/** Needed to compile on RH 4.X for some reason **/

#ifndef MAP_FAILED
#define MAP_FAILED -1
#endif

int wavfd;

/** command line options - stored as globals **/
int verbose = 0;             /* -v */ 
int destructive = 0;         /* -d */
double dur_in = 0.0;         /* -T */
double dur_out = 0.0;        /* -t */
double lvl_in = 0.0;         /* -L */
double lvl_out =0.0;         /* -l */



void usage(char *progname);
void fade_wav_file(char *filename);

int main(int argc,char *argv[])
{
   int c;

   while ((c = getopt(argc, argv, "vdT:t:L:l:")) != EOF) 
   {
      switch(c) 
      {
        case 'v':
          verbose = 1;
          printf("%s version %s 1999\n",argv[0],VERSION);
          break;
        case 'd':
          destructive = 1;
          break;
        case 't':
          dur_out = atof(optarg);
          if (verbose)
            printf("Fade-out %f sec.\n", dur_out);
          break;
        case 'l':
          lvl_out = atof(optarg);
          if (verbose)
            printf("Fade-out to level %f. %%.\n", lvl_out);
          break;
        case 'T':
          dur_in = atof(optarg);
          if (verbose)
            printf("Fade-in %f sec.\n", dur_in);
          break;
        case 'L':
          lvl_in = atof(optarg);
          if (verbose)
            printf("Fade-in from level %f. %%.\n", lvl_in);
          break;
          
          
        default:
          usage(argv[0]);
          exit(1);
      }
   }

   if ((optind + 1) > argc)
   {
     usage(argv[0]);
     exit(1);
   }

   for (c = optind; c < argc; c++)
   {
     fade_wav_file(argv[c]);
   }
   
   return(0);
}


void fade_wav_file(char *filename)
{      
  char *inwavbuf;
  short *current;
  short *audio;
  short *audioend;
  WAVE_HEADER *wav_info;
  int wavfd;
  struct stat input_fstat;
  size_t increment;
  double x;
  unsigned long n_in, n_out;
  
  printf("processing %s\n",filename);
  
  wavfd = open(filename,O_RDWR,0600);
  if (wavfd == -1)
  {
    printf("Error: open() %s\n",strerror(errno));
    return;
  }
  
  if (fstat(wavfd,&input_fstat) != 0)
  {
    printf("Error: fstat() %s\n",strerror(errno));
    if (close(wavfd) != 0)
    {
      printf("Error: close() %s\n",strerror(errno));
      exit(1);
    }
    return;
  }
  
  if (input_fstat.st_size < sizeof(WAVE_HEADER))
  {
    printf("Error: %s not large enough to hold a .wav file header!\n",
           filename);
    return;
  }
  
  inwavbuf = mmap(NULL,input_fstat.st_size,
                  PROT_READ | PROT_WRITE,MAP_SHARED,wavfd,0);
  if (inwavbuf == MAP_FAILED)
  {
    printf("Error: mmap() %s\n",strerror(errno));
    return;
  }
  
  audio = (short *)validate_wav_header(inwavbuf,verbose);
  
  if (audio == NULL)
  {
    printf("Error: %s not in suitable .wav format for this program\n",
           filename);
    return;
  }
  
  wav_info = (WAVE_HEADER *)inwavbuf;
  
  increment = wav_info->wBitsPerSample / 8;
  
  if (wav_info->nDataBytes != input_fstat.st_size - sizeof(WAVE_HEADER))
  {
    printf("   Warning: file length does not match length in header.\n");
    
    if (destructive)
    {
      printf("   --> fixing length to corrrect value\n");
      wav_info->nDataBytes = input_fstat.st_size - sizeof(WAVE_HEADER);
    }
  }
  
  current = audio;
  audioend = (short *)((char *)audio + wav_info->nDataBytes);
  
  if (destructive)
  {
    if (verbose)
      printf("    fading now...\n");

    /* convert % to {0..1} */
    lvl_in = lvl_in / 100;
    lvl_out = lvl_out / 100;
    /* convers seconds to samples */
    n_in = wav_info->nSamplesPerSec * dur_in;
    n_out = wav_info->nSamplesPerSec * dur_out;
    
    while (current < audioend)
    {
      x = 1.0;
      if( (current - audio) < n_in )
        x = lvl_in + (1-lvl_in) * (float)((float)(current - audio) / (float)n_in);
      if( (audioend - current) < n_out )
        x = lvl_out + (1-lvl_out) * (float)((float)(audioend - current) / (float)n_out);
#ifdef DEBUG
      if((x < 0.9999) && (current-audio)%1000 == 0)
        printf("[%d] x=%f\n", (current-audio), x);
#endif      
      *current = (int )((double )*current * x);
      current++;
         }
  }
  
  munmap(inwavbuf,input_fstat.st_size);
  
  if (close(wavfd) != 0)
  {
    printf("Error: close() %s\n",strerror(errno));
    exit(1);
  }
  
}

void usage(char *progname)
{
   printf("%s: version %s\n",progname,VERSION);
   printf("%s [-t threshold] [-v] [-d] [-t time] [-l level] [-T time] [-L level] file.wav [file2.wav ...]\n",
      progname);
   printf("   -v verbose, and version info\n");
   printf("   -d destructive (actually alter the .wav!)\n");
   printf("   -t fade-out duration / seconds (dfl=0)\n");
   printf("   -l fade-out target level / %% (dfl=0)\n");
   printf("   -T fade-in duration / seconds (dfl=0)\n");
   printf("   -L fade-in start level / %% (dfl=0)\n");
}
