/*
 * WAVLEVEL - a batch processing WAV file normalizer
 *
 * This work is based on WAVNORM by duncan@zog.net.au
 * Different algorithms have been added by
 * frederik.arbeiter@gmx.net
 * 
 */

#define WLVERSION "0.2"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <math.h>
#include <limits.h>

#include "wavfuncs.h"

#define DEBUG

#define MAXAMPLITUDE 32766 
#define ENERGY_NORMAL 3300
#define VOLUME_NORMAL 10000
#define LOGVOL_NORMAL log10(VOLUME_NORMAL)
#define MAXFILES 100

typedef short AUDIODATA;

/** access multi channel audio data */
#define VALUE(start,channel,index) start[(index << 1)+channel]

/** store statistics */
typedef struct
{
  AUDIODATA min, max;
  double sum_pp;
  double sum_energy;
  double sum_log_pp;
  unsigned long ncycles;
  long volume_avr; 
  long volume_max;
  double energy_avr;
  double energy_max;
  double logvol_avr;
  double logvol_max;
} SCANRESULTS;

/**complete data of 1 wav file */
typedef struct
{
  char filename[100];
  int wavfd;
  char *inwavbuf;
  struct stat input_fstat;
  WAVE_HEADER *wav_info;  

  AUDIODATA *audio;

  SCANRESULTS scanres;
} WAV_ENTITY;

/** 4 leveling methods available.
    NORM: use maximal dynamic available
    AMPL: scale to a "normal" average peak to peak value. Normal is AMPLITUDE_NORMAL in
          single file mode or maximum possible in batch mode
    LAMPL: same as AMPL, but uses log10(amplitude) as a measure
    POWER: same as AMPL, but uses energy (amplitude*frequency) values instead of peak to peak span as measure
*/
enum METHODLIST {NORM, POWER, AMPL, LAMPL};

/** Needed to compile on RH 4.X for some reason **/
#ifndef MAP_FAILED
#define MAP_FAILED -1
#endif

/** our WAV file(s) */
WAV_ENTITY* wfiles[MAXFILES];

/** command line options - stored as globals **/
int verbose = 0;                /* -v */ 
int destructive = 0;            /* -d */
double threshold = -1.0;        /* -t */
double ratio = 1.0;             /* -r */
enum METHODLIST method = AMPL;  /* -m */
double override_norm = 0.0;     /* -o */
int use_override = 0;           
int batchmode = 0;              /* -b */
int curious = 0;                /* -c */



void usage(char *progname);
void open_wav_file(WAV_ENTITY* w);
void close_wav_file(WAV_ENTITY* w);
void normalise_wav_file(char* filename);
void analyze_wav_file(WAV_ENTITY* w);
void relevel_wav_file(WAV_ENTITY* w, double x);
long max(long a, long b);





int main(int argc,char *argv[])
{
   int c;
   double maxdyn, dyn, pp, carry_pp;
   double x = 1.0;
   double carry_norm, global_norm;

   while ((c = getopt(argc, argv, "vdtbc:r:o:m:")) != EOF) 
   {
      switch(c) 
      {
        case 'v':
          verbose = 1;
          printf("wavfuncs version %s duncan@zog.net.au 1999\n",VERSION);           
          printf("%s version %s frederik.arbeiter@gmx.net 2000\n",argv[0],WLVERSION);
          break;
        case 'd':
          destructive = 1;
          break;        
        case 't':
          threshold = atof(optarg);
          if (verbose)
            printf("Using threshold of ratio of %f\n",threshold);
          break;
        case 'b':
          batchmode = 1;
          break;
        case 'c':
          curious = 1;
          verbose =1; /* If we take the time to calculate, we should also tell the user ... */
          break;
        case 'r':
          ratio = atof(optarg);
          break;
        case 'o':
          override_norm = atof(optarg);
          use_override = 1;
          break;
          
        case 'm':
          switch(optarg[0])
          {
            case 'n':
              method = NORM;
              break;
            case 'p':
              method = POWER;
              break;
            case 'a':
              method = AMPL;
              break;
            case 'l':
              method = LAMPL;
              break;
            default:
              fprintf(stderr, "invalid  method flag `%c` passed\n", optarg[0]);
              usage(argv[0]);
              exit(-1);
              break;
          }
          break;

        default:
          usage(argv[0]);
          exit(1);
      }
   }
   
   if ((optind + 1) > argc)
   {
      usage(argv[0]);
      exit(1);
   }

   if(method == NORM)
     /* behave like wavnorm.c */
   {
      for(c=optind; c<argc; c++)
      {
        normalise_wav_file(argv[c]);
      }
      return 0;
   }
   

   if((argc-optind) > MAXFILES)
   {
     fprintf(stderr, "Sorry, can only handle %d files at a time!\n", MAXFILES);
     exit(-1);
   }
   
   
   for(c=optind; c<argc; c++)
   {
     wfiles[c-optind] = (WAV_ENTITY*) malloc(sizeof(WAV_ENTITY));
     strcpy(wfiles[c-optind]->filename, argv[c]);
     
     open_wav_file(wfiles[c-optind]);
     analyze_wav_file(wfiles[c-optind]);
     close_wav_file(wfiles[c-optind]);
   }

   maxdyn = 0.0;
   carry_pp = 0.0;
   carry_norm = 0.0;
   global_norm = 0.0;
   
   
   for(c=0; c<argc-optind; c++)
   {
     pp = 2*max(wfiles[c]->scanres.max, -wfiles[c]->scanres.min);
     switch(method)
     {
       case POWER:
         dyn =  pp / wfiles[c]->scanres.energy_avr;
         if(dyn > maxdyn)
         {
           maxdyn = dyn;
           carry_pp = pp;
           carry_norm = wfiles[c]->scanres.energy_avr;
         }
         break;
       case AMPL:
         dyn = pp / wfiles[c]->scanres.volume_avr;
         if(dyn > maxdyn)
         {
           maxdyn = dyn;
           carry_pp = pp;
           carry_norm = wfiles[c]->scanres.volume_avr;
         }
         break;
       case LAMPL:
         dyn = pp / wfiles[c]->scanres.logvol_avr;
         if(dyn > maxdyn)
         {
           maxdyn = dyn;
           carry_pp = pp;
           carry_norm = wfiles[c]->scanres.logvol_avr;
         }
         break;
         
       default:
         printf("Invalid method ?\n");
         exit(-1);
     }
   }
   global_norm = (double) (MAXAMPLITUDE*2-2)/carry_pp * carry_norm;

   if(verbose)
   {
     printf("\nglobal maximum dynamics ratio is %.3f\n", maxdyn);
     printf("at normvalue %.3f and value range span = %.3f.\n", carry_norm, carry_pp);
     printf("global normvalue calculated to %.3f.\n\n", global_norm);
   }
   
   if(use_override)
   {
     global_norm = override_norm;
     if(verbose)
       printf("global normvalue used is %.3f due to user override option\n\n", global_norm);
   }
   
   for(c=optind; c<argc; c++)
   {
     switch(method)
     {
       case POWER:
         x = global_norm / wfiles[c-optind]->scanres.energy_avr;
         break;
       case AMPL:
         x = global_norm / wfiles[c-optind]->scanres.volume_avr;
         break;
       case LAMPL:
         x = global_norm / wfiles[c-optind]->scanres.logvol_avr;
         break;
       default:
         x = 1.0;
         break;
     }
     
     open_wav_file(wfiles[c-optind]);
     relevel_wav_file(wfiles[c-optind], x);
     close_wav_file(wfiles[c-optind]);
     
   }
   
   for(c=optind; c<argc; c++)
   {
     free(wfiles[c-optind]);
   }
   
   return 0;
}


long  max(long a, long b)
{
  if(a>b)
    return a;
  return b;
}


int asgn(long arg)
/* asymetric signum */
{
  if(arg > 0)
    return 1;
  else
    return -1;
}


int find_cycle(AUDIODATA* data, int chn, unsigned long data_n, unsigned long startsearch,
               unsigned long* cycle_begin, unsigned long* cycle_end)
/* return index to begin and end of next cycle (complete wave begining and ending with zero
   value) in audio data
   returns -1 on failure/no complete cycle left */
{
  int sgn_carry;
  *cycle_begin = startsearch;
  *cycle_end = startsearch;
  sgn_carry = asgn(VALUE(data,chn,*cycle_begin));
  
  /* searching begin */
  while(asgn(VALUE(data,chn,*cycle_begin)) == sgn_carry)
  {
    // printf("b: [%d]=%d / %d\n",*cycle_begin, asgn(VALUE(data,chn,*cycle_begin)), sgn_carry);
    if(*cycle_begin < data_n-1)
      (*cycle_begin)++;
    else
      return -1; /* can't find beginning within data field */
  }
  
  *cycle_end = *cycle_begin;
  sgn_carry = asgn(VALUE(data,chn,*cycle_end));
  
  /* searching next zero */
  while(asgn(VALUE(data,chn,*cycle_end)) == sgn_carry)
  {
    // printf("m: [%d]=%d / %d\n",*cycle_end, asgn(VALUE(data,chn,*cycle_end)), sgn_carry);
    if(*cycle_end < data_n-1)
      (*cycle_end)++;
    else
      return -1; /* can't proceed to next halfwave */
  }
  
  sgn_carry = asgn(VALUE(data,chn,*cycle_end));
  
  /* searching end */
   while(asgn(VALUE(data,chn,*cycle_end)) == sgn_carry)
   {
     // printf("e: [%d]=%d / %d\n",*cycle_end, (VALUE(data,chn,*cycle_end)), sgn_carry);
     if(*cycle_end < data_n-1)
       (*cycle_end)++;
     else
       return -1;
   }

   (*cycle_end)--;
   //printf("find: %d, %d\n", *cycle_begin, *cycle_end);
   
   
   return 1;
   
}

void open_wav_file(WAV_ENTITY* w)
{
   size_t increment;

   printf("processing %s\n",w->filename);

   w->wavfd = open(w->filename,O_RDWR,0600);
   if (w->wavfd == -1)
   {
      printf("Error: open() %s\n",strerror(errno));
      return;
   }

   if (fstat(w->wavfd, &(w->input_fstat)) != 0)
   {
      printf("Error: fstat() %s\n",strerror(errno));
      if (close(w->wavfd) != 0)
      {
         printf("Error: close() %s\n",strerror(errno));
         exit(1);
      }
      return;
   }

   if (w->input_fstat.st_size < sizeof(WAVE_HEADER))
   {
      printf("Error: %s not large enough to hold a .wav file header!\n",
         w->filename);
      return;
   }

   w->inwavbuf = mmap(NULL, w->input_fstat.st_size,
      PROT_READ | PROT_WRITE,MAP_SHARED, w->wavfd,0);
   if (w->inwavbuf == MAP_FAILED)
   {
      printf("Error: mmap() %s\n",strerror(errno));
      return;
   }

   w->audio = (short *)validate_wav_header(w->inwavbuf, verbose);

   if (w->audio == NULL)
   {
      printf("Error: %s not in suitable .wav format for this program\n",
         w->filename);
      return;
   }

   w->wav_info = (WAVE_HEADER *) w->inwavbuf;

   increment = w->wav_info->wBitsPerSample / 8;

   if (w->wav_info->nDataBytes != w->input_fstat.st_size - sizeof(WAVE_HEADER))
   {
      printf("   Warning: file length does not match length in header.\n");

      if (destructive)
      {
         printf("   --> fixing length to corrrect value\n");
         w->wav_info->nDataBytes = w->input_fstat.st_size - sizeof(WAVE_HEADER);
      }
   }
}



void analyze_wav_file(WAV_ENTITY* w)
{
  unsigned long index, cycle_end, nSamples, pp;
   AUDIODATA hi, lo;
   long t;
   double en, lpp;
   
   if (verbose)
      printf("  analyzing volume statistics...");

   w->scanres.min = MAXAMPLITUDE;
   w->scanres.max = -MAXAMPLITUDE;
   w->scanres.sum_pp = 0.0;
   w->scanres.sum_energy = 0.0;
   w->scanres.sum_log_pp = 0.0;
   w->scanres.ncycles = 0;
   w->scanres.volume_avr = 0;
   w->scanres.volume_max = 0;
   w->scanres.logvol_avr = 0.0;
   w->scanres.logvol_max = 0.0;
   
   nSamples = w->wav_info->nDataBytes/4; /* we allowed 16bit stereo samples  only... */

   index = 0;
   while (index < nSamples)
   {
     if(0 < find_cycle(w->audio, 0, nSamples, index, &index, &cycle_end))
     {
       lo = MAXAMPLITUDE;
       hi = -MAXAMPLITUDE;
       t = cycle_end-index;
       
       for(; index <= cycle_end; index++)
       {
#ifdef DEBUG
         //printf("audio[%d] = %d, lo=%d, hi=%d\n", index, VALUE(audio,0,index), lo, hi);
#endif         
         if(VALUE(w->audio,0,index) > hi)
           hi = VALUE(w->audio,0,index);
         if(VALUE(w->audio,0,index) < lo)
           lo = VALUE(w->audio,0,index);
       }
       
       if(lo < w->scanres.min)
         w->scanres.min = lo;
       if(hi > w->scanres.max)
         w->scanres.max = hi;

       pp = hi - lo;
       w->scanres.sum_pp += pp;
       if(pp > w->scanres.volume_max)
         w->scanres.volume_max = pp;

       if(curious || (method == LAMPL))
       {
         lpp = log10(pp);
         w->scanres.sum_log_pp += lpp;
         if(lpp > w->scanres.logvol_max)
           w->scanres.logvol_max = lpp;
       }
       
         
       if(curious || (method == POWER))
       {
         en = ((double)pp/(double)t);
         w->scanres.sum_energy += en;
         if(en > w->scanres.energy_max)
           w->scanres.energy_max = en;
       }
         
       w->scanres.ncycles++;

       
       
       //w->scanres.volume_distrib[VSTEPS*pp/65536]++;
       //w->scanres.energy_distrib[VSTEPS*pp/t/10000]++;
       
#ifdef DEBUG
       //printf("\nindex=%d, hi=%d, lo=%d, pp = %d\n",index, hi, lo, pp);
#endif
       index--;
     }
     else
     {
       //printf("find_cycle() failed\n.");
       break;
     }
   }
   
   w->scanres.volume_avr = w->scanres.sum_pp / w->scanres.ncycles;
   w->scanres.logvol_avr = w->scanres.sum_log_pp / w->scanres.ncycles;
   w->scanres.energy_avr = w->scanres.sum_energy / w->scanres.ncycles;
   
   if(verbose)
   {
     printf("\n");
     printf("    min. sample=%d, max. sample=%d\n", w->scanres.min, w->scanres.max);
     printf("    avr. pp=%ld, max. pp=%ld in %ld cycles\n", w->scanres.volume_avr, w->scanres.volume_max, w->scanres.ncycles);
     if(curious || (method == LAMPL))
       printf("    avr. lpp=%f, max. lpp=%f\n", w->scanres.logvol_avr, w->scanres.logvol_max);
     if(curious || (method == POWER))
       printf("    avr. en=%f, max. en=%f\n", w->scanres.energy_avr, w->scanres.energy_max);
   }

   return;
   
   
}

void relevel_wav_file(WAV_ENTITY* w, double x)
{
  AUDIODATA* current;
  AUDIODATA* audioend;
  AUDIODATA offset=0;
  
  int loudest = 0;
  double loudestd;
  double loudr;
  
  
  current = w->audio;
  audioend = (short *)((char *)w->audio + w->wav_info->nDataBytes);

  offset = - (w->scanres.max + w->scanres.min) / 2;
  loudestd = fabs((double )loudest);
  loudr =  x * ratio;
  if (verbose)
  {
    //printf("\n");
    printf("    offset is %d \n", offset);
    printf("    amplification is %f\n", loudr);
    printf("    max = %d -> %d, min = %d -> %d\n",
           w->scanres.max, (int) ((w->scanres.max+offset)*loudr),
           w->scanres.min, (int) ((w->scanres.min+offset)*loudr));
    if(w->scanres.energy_avr > 0.001)
      printf("    avr energy %.3f -> %.3f\n",
             w->scanres.energy_avr, loudr*w->scanres.energy_avr);
    if(w->scanres.volume_avr > 0.001)
      printf("    avr pp. %ld -> %.3f\n",
             w->scanres.volume_avr, loudr*w->scanres.volume_avr);
    if(w->scanres.logvol_avr > 0.001)
      printf("    avr log10(pp) %.3f -> %.3f\n",
             w->scanres.logvol_avr, loudr*w->scanres.logvol_avr);

  }

  if((loudr * (w->scanres.max) > MAXAMPLITUDE) || (loudr * (w->scanres.min) < -MAXAMPLITUDE))
  {
    printf("BUG! Value over board!\n");
    printf("switching off destructivity now!\n");
     destructive = 0; 
   } 
  
  if (destructive)
  {
    if (loudr > threshold)
    {
      if (verbose)
        printf("    scaling with ratio %f now...\n",loudr);
      
      /** we go backwards through the audio now, as presumably the
          last bytes we read are more likely to be buffered, so
          hopefully this is a bit faster? **/
      
      while (current < audioend)
      {
        *current = (int )((double )(*current) * loudr);
        current++;
      }
    }
    else
    {
      if (verbose)
        printf("   threshold not hit, file not altered\n");
    }
  }
}



void close_wav_file(WAV_ENTITY* w)
{
  munmap(w->inwavbuf, w->input_fstat.st_size);
  
  if (close(w->wavfd) != 0)
  {
    printf("Error: close() %s\n",strerror(errno));
    exit(1);
  }
}



void normalise_wav_file(char *filename)
{      
   char *inwavbuf;
   short *current;
   short *audio;
   short *audioend;
   WAVE_HEADER *wav_info;
   int wavfd;
   struct stat input_fstat;
   size_t increment;
   int loudest = 0;
   double loudestd;
   double total;
   double loudr;

   printf("processing %s\n",filename);

   wavfd = open(filename,O_RDWR,0600);
   if (wavfd == -1)
   {
      printf("Error: open() %s\n",strerror(errno));
      return;
   }

   if (fstat(wavfd,&input_fstat) != 0)
   {
      printf("Error: fstat() %s\n",strerror(errno));
      if (close(wavfd) != 0)
      {
         printf("Error: close() %s\n",strerror(errno));
         exit(1);
      }
      return;
   }

   if (input_fstat.st_size < sizeof(WAVE_HEADER))
   {
      printf("Error: %s not large enough to hold a .wav file header!\n",
         filename);
      return;
   }

   inwavbuf = mmap(NULL,input_fstat.st_size,
      PROT_READ | PROT_WRITE,MAP_SHARED,wavfd,0);
   if (inwavbuf == MAP_FAILED)
   {
      printf("Error: mmap() %s\n",strerror(errno));
      return;
   }

   audio = (short *)validate_wav_header(inwavbuf,verbose);

   if (audio == NULL)
   {
      printf("Error: %s not in suitable .wav format for this program\n",
         filename);
      return;
   }

   wav_info = (WAVE_HEADER *)inwavbuf;

   increment = wav_info->wBitsPerSample / 8;

   if (wav_info->nDataBytes != input_fstat.st_size - sizeof(WAVE_HEADER))
   {
      printf("   Warning: file length does not match length in header.\n");

      if (destructive)
      {
         printf("   --> fixing length to corrrect value\n");
         wav_info->nDataBytes = input_fstat.st_size - sizeof(WAVE_HEADER);
      }
   }

   current = audio;

   if (verbose)
      printf("   finding highest level...");

   audioend = (short *)((char *)audio + wav_info->nDataBytes);

   while (current < audioend)
   {
      if (abs(*current) > loudest)
      {
         if (verbose)
         {
            printf(".");
            fflush(stdout); /* so the dot appears quickly! */
            total = (double )abs(*current);
         }
         loudest = abs(*current);
         
      }
      current++;
   }


   loudestd = fabs((double )loudest);
   loudr =  SHRT_MAX / loudestd * ratio;
   if (verbose)
   {
      printf("\n");
      printf("    scaling to ratio %f of max\n",ratio);
      printf("    normalise ratio is %f\n",loudr);
   }

   if (destructive)
   {
      if (loudr > threshold)
      {
         if (verbose)
            printf("    normalising with ratio %f now...\n",loudr);

         /** we go backwards through the audio now, as presumably the
             last bytes we read are more likely to be buffered, so
             hopefully this is a bit faster? **/
  
         while (current > audio)
         {
            *current = (int )((double )*current * loudr);
            current--;
         }
      }
      else
      {
         if (verbose)
            printf("   threshold not hit, file not altered\n");
      }
   }

   munmap(inwavbuf,input_fstat.st_size);

   if (close(wavfd) != 0)
   {
      printf("Error: close() %s\n",strerror(errno));
      exit(1);
   }
}



void usage(char *progname)
{
  printf("\nwavfuncs version %s duncan@zog.net.au 1999\n",VERSION);           
  printf("%s version %s frederik.arbeiter@gmx.net 2000\n", progname ,WLVERSION);
  printf("%s [OPTIONS] file.wav [file2.wav ...]\n",
      progname);
  printf("OPTIONS:\n");
  printf("   -v verbose, and version info\n");
  printf("   -d destructive (actually alter the .wav!)\n");
  printf("   -t <threshold> threshhold ratio under which dont alter\n");
  printf("   -r <ratio>     ratio of max possible to scale (default 1.0)\n");
  printf("   -o <normlevel> override normlevel\n");
  printf("   -c curiosity. Calculate parameters for ALL methods, just for education.\n");
  printf("   -m <method> {n(ormalize), p(ower), a(mplitude) l(ogarithmic amplitude)} (default a)\n");
}

